"""
Provides the GUI for editing OMFITnamelist instances in the tree when they are double-clicked.
"""

defaultVars(nml=None, misc_label='Other')
if nml is None:
    nml = OMFITnamelist('new_nml')

try:
    OMFITx.TitleGUI(relativeLocations(nml)['OMFITlocationName'][-1])
except Exception:
    OMFITx.End()

MainScratch.setdefault(id(nml), {})  # Set up a place to track the status of the filters
misc_tab_opened = False  # Track whether the misc tab has been opened yet so we don't create extra filter entries in it.

for kn, n in nml.items():
    if kn.startswith('__'):  # Hidden item: do nothing
        continue

    if isinstance(n, omfit_classes.namelist.NamelistName):  # Make a new tab for a NamelistName and load its children
        OMFITx.Tab(kn)
        OMFITx.Entry("MainScratch[%d]['%s_filter']" % (id(nml), kn), lbl='Filter:', default='', updateGUI=True)
        OMFITx.Separator()
        nml_loc = relativeLocations(nml[kn])['OMFITlocationName'][-1]

        for k, v in n.items():
            if k[:2] == '__':  # Don't display entries for hidden children
                continue
            if MainScratch[id(nml)]['%s_filter' % kn].lower() in k.lower():  # Make sure item name passes filter
                OMFITx.Entry(nml_loc + "['%s']" % k, lbl=k)  # Make an entry for an item within a NamelistName

    else:
        OMFITx.Tab(misc_label)
        if not misc_tab_opened:
            # Create the misc tab's filter (first time only)
            OMFITx.Entry("MainScratch[%d]['%s_filter']" % (id(nml), misc_label), lbl='Filter:', default='', updateGUI=True)
            OMFITx.Separator()
            misc_tab_opened = True

        try:
            nml_loc = relativeLocations(nml)['OMFITlocationName'][-1]
        except Exception as _excp:
            printe(repr(E))
            printe('Failed to parse %s' % kn)
            continue

        if MainScratch[id(nml)]['%s_filter' % misc_label] in kn.lower():  # Make sure item name passes filter
            OMFITx.Entry(nml_loc + "['%s']" % kn, lbl=kn)  # Make an entry for a top-level non-NamelistName item


def cleanup(topGUI):
    if id(nml) in MainScratch:
        del MainScratch[id(nml)]
    OMFITx._clearClosedGUI(topGUI)


OMFITx._aux['topGUI'].protocol('WM_DELETE_WINDOW', lambda topGUI=OMFITx._aux['topGUI']: cleanup(topGUI))



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

import os
from omfit_classes.omfit_json import OMFITjson


class _OMFITclusters(OMFITjson):
    """
    This class finds the optimal batch submission configuration for a given system.
    Designed to work in conjunction with OMFITx.job_array.

    Reads in the configuration of through the `omfit_clusters.json` file.
    Some notes on the json file:
    - `partitions_decision` and `qos_decision` describe the qualities of `partitions` and `qos`.
      The array of length 4 contains:
        - Minimum time for a queue: Used to indicate that another queue with a shorter maximum wall time exists.
                                    This allows the queue with the shortest maximum wall time that can fit the job
                                    is used.
        - Maximum time for a queue: Upper limit on wall time for queue.

        - Maximum number of concurrent cores: Makes sure that the queue can provide the resources requested by the job.

        - Total virtual memory provided by a single node: Used to calculate the amount of memory per core for the job.

    """

    def __init__(self):
        OMFITjson.__init__(self, os.path.split(__file__)[0] + os.sep + 'omfit_clusters.json')

    def get_cluster_config(self, cluster):
        """
        Returns configuration of selected cluster.

        :param cluster: Name of the target cluster

        :return: Dictionary of cluster meta information.
        """
        try:
            cluster_config = self[cluster]
        except KeyError:
            raise NotImplementedError(
                f"""
                    The cluster {cluster} is not yet implemented.\n
                    Please consider adding its configuration to omfit_clusters.json
                    """
            )
        return cluster_config

    def get_cpus_per_node(self, cluster, constraint=None):
        """
        Returns maximum allowed concurrent gpu cores.
        Allows assignment to multiple nodes.

        :param cluster: Name of the target cluster

        :param wall_time: Wall time in hours

        :return: Maximum allowed concurrent gpu cores

        """
        cluster_config = self.get_cluster_config(cluster)
        if constraint is None:
            return cluster_config["cpus_per_node"]
        else:
            try:
                constraint_index = cluster_config["constraints"].index(constraint)
                return cluster_config["cpus_per_node"][constraint_index]
            except ValueError:
                raise NotImplementedError(
                    f"""
                    The cluster {self.cluster} does not have the contraint
                    {constraint} you requested implemented.
                    lease consider adding this constraint to omfit_clusters.json
                    """
                )

    def get_max_distriubted_cores_from_wall_time(self, cluster, wall_time):
        """
        Returns maximum allowed concurrent cpu cores.
        Allows assignment to multiple nodes.

        :param cluster: Name of the target cluster

        :param wall_time: Wall time in hours

        :return: Maximum allowed concurrent cpu cores

        """
        cluster_config = self.get_cluster_config(cluster)
        max_distributed_cores = 0
        for iqueue, queue_select in enumerate(["partitions", "qos"]):
            if cluster_config[queue_select] is not None:
                for queue, decision in zip(cluster_config[queue_select], cluster_config[queue_select + "_decision"]):
                    if wall_time > decision['t_min'] and wall_time <= decision['t_max'] and decision['max_cpus'] > max_distributed_cores:
                        max_distributed_cores = decision['max_cpus']
        return max_distributed_cores

    def get_batch_simple(
        self, cluster, wall_time, ntasks, cpu_per_task, mem_per_task=None, partition=None, qos=None, constraint=None, vmem_per_node=None
    ):
        """
        Generates the Slurm commands needed for OMFITx.job_array

        :param cluster: Name of the cluster (lowercase)

        :param wall_time: Wall time as a floating point number in hours

        :param ntasks: Number of MPI tasks every job in the job array uses

        :param cpu_per_task: Number of cpus allocated for each MPI task

        :param mem_per_task: Memory requirements of task [GB]

        :param partition: Force the use of a certain partition. The routine will look up the most suitable partition otherwise.

        :param qos: Force the use of a certain quality-of-service. The routine will look up the most suitable qos otherwise.

        :param constraint: Additional constraint for Slurm. Used for example by NERSC.

        :param vmem_per_node: When specifying quality of service/partition manually you also need to provide the amount of vmem a node has [GB].

        :return: Returns dictionary that can be passed on to OMFITx.job_array via the ** operator
        """
        cluster_config = self.get_cluster_config(cluster)
        for mandatory_arg in cluster_config["mandatory_arguments_for_sbatch_simple"]:
            if locals()[mandatory_arg] is None:
                raise ValueError(f"For {cluster} {mandatory_arg} is a mandatory argument!")
        omfit_job_array_config = {}
        omfit_job_array_config["partition"] = ""
        queue_select_success = [False, False]
        cpu_count = ntasks * cpu_per_task
        query = []
        if partition is None:
            query.append("partitions")
        else:
            if "--partition=" not in partition:
                raise ValueError("Please provide the partition keyword in the form --partition==<your partition>")
            omfit_job_array_config["partition"] += partition
        if qos is None:
            query.append("qos")
        else:
            if "--qos=" not in qos:
                raise ValueError("Please provide the qos keyword in the form --qos=<your qos>")
            if not omfit_job_array_config["partition"].endswith(" ") and len(omfit_job_array_config["partition"]) > 0:
                omfit_job_array_config["partition"] += " "
            omfit_job_array_config["partition"] += qos
        for iqueue, queue_select in enumerate(query):
            if cluster_config[queue_select] is not None:
                # Note that omfit.job_array only takes "partition" as an argument and not qos
                if not omfit_job_array_config["partition"].endswith(" ") and len(omfit_job_array_config["partition"]) > 0:
                    omfit_job_array_config["partition"] += " "
                for queue_meta, decision in zip(cluster_config[queue_select], cluster_config[queue_select + "_decision"]):
                    if wall_time > decision['t_min'] and wall_time <= decision['t_max'] and decision['max_cpus'] >= cpu_count:
                        omfit_job_array_config["partition"] += queue_meta
                        queue_select_success[iqueue] = True
                        if vmem_per_node is None or decision['vmem_per_node'] < vmem_per_node:
                            vmem_per_node = decision['vmem_per_node']
                        break
            else:
                queue_select_success[iqueue] = True
        if vmem_per_node == -1:
            raise ValueError(
                f"""
                    Could not automatically detect memory per node. You might need to add
                    extra configuration data for the cluster/partition/qos you requested.
                    Requested cluster/partition/qos {cluster} {partition} {qos}
                """
            )
        omfit_job_array_config["partition_flag"] = ""
        if constraint is None and cluster_config["constraints"] is not None:
            # Use primary constraint as default
            omfit_job_array_config["partition"] += " " + cluster_config["constraints"][0]
            constraint = cluster_config["constraints"][0]
        elif constraint is not None:
            omfit_job_array_config["partition"] += " " + constraint
        omfit_job_array_config["nproc_per_task"] = ntasks
        hours = int(wall_time)
        minutes = int(wall_time * 60 - hours * 60)
        seconds = int(wall_time * 3600 - hours * 3600 - minutes * 60)
        omfit_job_array_config["job_time"] = "{0:d}:{1:02d}:{2:02d}".format(hours, minutes, seconds)
        omfit_job_array_config["batch_type"] = cluster_config["batch_type"]
        n_max_cpus = self.get_cpus_per_node(cluster, constraint)
        # Since omfitx.job_array do not specify
        # the number of nodes explicitly, we will be sharing nodes.
        # Therefore, the memory should be limited to the amount of memory per cpu on the system
        if mem_per_task is not None:
            if vmem_per_node < mem_per_task:
                raise ValueError(f"Insufficient memory for on a node {vmem_per_node} for this task {mem_per_task}")
            else:
                omfit_job_array_config["memory"] = f"{int(mem_per_task / cpu_per_task * 0.95 * 1024)}M"
        else:
            omfit_job_array_config["memory"] = f"{int(vmem_per_node / n_max_cpus * 0.95 * 1024)}M"
        return omfit_job_array_config


OMFITclusters = _OMFITclusters()
OMFITclusters.load()

__all__ = ["OMFITclusters"]
if __name__ == "__main__":
    OMFITclusters.get_batch_simple("perlmutter", 2.0, 1, 1, constraint="--constraint=cpu")



from setuptools import setup
import os

# ======================
# Install `omfit_classes` package
# ======================

os.chdir(os.path.split(os.path.realpath(__file__))[0])

with open('omfit_classes/version') as f:
    version = f.read().strip() + '.2'

with open('README.rst') as f:
    long_description = '\n' + '\n'.join(f.read().strip().split('\n')[2:])

package_data = ['../LICENSE.txt', '../README.rst', '*']
for root, dir, files in os.walk("omfit_classes"):
    if root == 'omfit_classes':
        continue
    if '__pycache__' in root:
        continue
    package_data.append(root[len('omfit_classes') + 1 :] + os.sep + '*')

setup(
    name='omfit_classes',
    version=version,
    description='Classes of OMFIT (One Modeling Framework For Integrated Tasks)',
    long_description=long_description,
    url='https://omfit.io',
    author='OMFIT developers',
    author_email="meneghini@fusion.gat.com",
    classifiers=['Programming Language :: Python :: 3', 'Operating System :: OS Independent', 'License :: OSI Approved :: MIT License'],
    keywords='integrated modeling plasma framework',
    packages=['omfit_classes'],
    package_dir={'omfit_classes': 'omfit_classes'},
    install_requires=[],
    extras_require={},
    package_data={'omfit_classes': package_data},
    license='MIT',
)



print('Loading widgets utility functions...')

from omfit_classes.utils_base import *
from omfit_classes.utils_math import array_info
from tkinter import Tk as tk
from tkinter import ttk
import tkinter.font as tkFont
import numpy as np
import pandas
import xarray
import collections
from uncertainties.unumpy import nominal_values, std_devs

# -------------
# Tk utils
# -------------
rightClick = 'Button-3'
middleClick = 'Button-2'
rightClickIndex = 3
middleClickIndex = 2

hide_ptrn = re.compile(r'^__.*__$')


def treeText(inv, strApex=False, n=-1, escape=True):
    """
    Returns the string that is shown in the OMFITtree for the input object `inv`

    :param inv: input object

    :param strApex: should the string be returned with apexes

    :param n: maximum string length

    :param escape: escape backslash characters

    :return: string representation of object `inv`
    """
    if isinstance(inv, str):
        if strApex:
            return short_str(repr(inv), n, escape)
        elif isinstance(inv, str):
            return short_str(repr(inv)[1:-1], n, escape)

    elif isinstance(inv, list):

        def listType(x):
            lst = []
            for k in x[:10]:
                if isinstance(k, str):
                    kr = repr(k)
                    if len(kr) > 12:
                        k = kr[:11] + '..'
                elif hasattr(k, '__iter__'):
                    kr = '<%s>' % k.__class__.__name__
                elif isinstance(k, (int, np.integer, float, np.float, complex, np.complex)):
                    kr = repr(k)
                else:
                    kr = '<object>'
                lst.append(kr)
            if len(x) > 10:
                lst.append('...')
            return lst

        return short_str('[%s]' % ', '.join(listType(inv)), n, escape)

    elif is_float(inv):
        return format(inv, '3.8g')

    elif isinstance(inv, np.ndarray):
        return array_info(inv)

    elif pandas is not None and isinstance(inv, pandas.DataFrame):
        ncol = len(inv.columns)
        if ncol > 5:
            return ', '.join(k for k in inv.columns[:5]) + '...'
        else:
            return ', '.join(k for k in inv.columns)

    elif pandas is not None and isinstance(inv, pandas.Series):
        return (', '.join(l for l in str(inv.describe()).split('\n')[:2] if '%' not in l)).replace('\n', '')

    elif xarray is not None and isinstance(inv, xarray.DataArray):
        return ', '.join('%s: %s' % (k, v) for k, v in zip(inv.dims, inv.shape))

    elif xarray is not None and isinstance(inv, xarray.Dataset):
        return ', '.join(['%s: %s' % (k, inv.dims[k]) for k in inv.dims])

    elif isinstance(inv, (dict, collections.abc.MutableMapping)):
        if hasattr(inv, 'dynaLoad') and inv.dynaLoad:
            return '--{\t?\t}--'

        vshow = [k for k in inv if not isinstance(k, str) or not re.match(hide_ptrn, k)]
        vhide = [k for k in inv if isinstance(k, str) and re.match(hide_ptrn, k)]
        values = '--{\t'
        if vshow:
            values += '%d' % len(vshow)
        if vshow and vhide:
            values += ' '
        if vhide:
            values += '(%d)' % len(vhide)
        values += '\t}--'

        for k in ['modifyOriginal', 'readOnly']:
            if hasattr(inv, k) and getattr(inv, k):
                values += ' ' + k + ' '

        return values

    elif inspect.isroutine(inv):
        if inv.__doc__ is not None:
            tmp = inv.__doc__.strip().split('\n')
            tmp = [line for line in map(lambda x: x.strip(), tmp) if line and not line.startswith(':') and not line.startswith('>')]
            if tmp:
                return short_str(' | '.join(tmp), n, escape)
        try:
            return short_str('(' + function_arguments(inv, [], True).replace('\n', '').replace('self,', '') + ')', n, escape)
        except TypeError:
            return short_str('built-in method', n, escape)

    else:
        return short_str(repr(inv), n, escape)


def ctrlCmd():
    return 'Control'
    # return 'Command'


def short_str(inv, n=-1, escape=True, snip='[...]'):
    """
    :param inv: input string

    :param n: maximum length of the string (negative is unlimited)

    :param escape: escape backslash characters

    :param snip: string to replace the central part of the input string

    :return: new string
    """
    l = len(inv)
    s = len(snip)
    if not l:
        inv = "\'\'"
    elif l > n and n > s + 1:
        nn = (n - s) / 2.0
        a = int(round(nn))
        b = int(nn)
        inv = inv[:a] + snip + inv[-b:]
    if escape:
        inv = re.sub(r'\\', r'\\\\', inv)
    return inv


def tkStringEncode(inv):
    r"""
    Tk requires ' ', '\', '{' and '}' characters to be escaped
    Use of this function dependes on the system and the behaviour can be
    set on a system-by-system basis using the OMFIT_ESCAPE_TK_SPACES environmental variable

    :param inv: input string

    :return: escaped string
    """
    if int(os.environ.get('OMFIT_ESCAPE_TK_SPACES', '1')) and len(inv):
        inv = re.sub(r'([ \\\{\}])', r'\\\1', treeText(inv, strApex=False, n=-1, escape=False))
    return inv


_defaultFont = {}


def OMFITfont(weight='', size=0, family='', slant=''):
    """
    The first time that OMFITfont is called, the system defaults are gathered

    :param weight: 'normal' or 'bold'

    :param size: positive or negative number relative to the tkinter default

    :param family: family font

    :return: tk font object to be used in tkinter calls
    """

    font = {'family': 'Helvetica', 'size': 11, 'weight': 'normal', 'slant': 'roman'}

    # save default font if not done yet
    if not len(_defaultFont):
        ttk_style = ttk.Style()
        # generate font
        f = tkFont.Font(font=ttk_style.lookup('TLabel', 'font'))
        _defaultFont.update(f.metrics())

    # make sure the default font has all the categories
    for key, val in list(font.items()):
        _defaultFont.setdefault(key, val)
    if not _defaultFont['family']:  # bad things happen if this somehow gets set to ''
        _defaultFont['family'] = 'Helvetica'
    if _defaultFont['size'] < 6:  # bad things happen if this somehow gets set to 0
        _defaultFont['size'] = 6
    # modifications to default font
    font = {}
    font['size'] = abs(_defaultFont['size']) + size
    if weight:
        font['weight'] = weight
    elif _defaultFont['weight']:
        font['weight'] = _defaultFont['weight']
    if slant:
        font['slant'] = slant
    elif _defaultFont['slant']:
        font['slant'] = _defaultFont['slant']
    if family:
        font['family'] = family
    elif _defaultFont['family']:
        font['family'] = _defaultFont['family']

    return (font['family'], font['size'], font['weight'], font['slant'])


def tk_center(win, parent=None, width=None, height=None, xoff=None, yoff=None, allow_negative=False):
    """
    Function used to center a tkInter GUI

    Note: by using this function tk does not manage the GUI geometry
    which is beneficial when switching between desktop on some window
    managers, which would otherwise re-center all windows to the desktop.

    :param win: window to be centered

    :param parent: window with respect to be centered (center with respect to screen if `None`)

    :width: set this window width

    :height: set this window height

    :param xoff: x offset in pixels

    :param yoff: y offset in pixels

    :param allow_negative: whether window can be off the left/upper part of the screen
    """
    win.update_idletasks()

    if parent is None:
        utils_tk._winfo_screen(reset=True)
        swidth, sheight, x0, y0 = screen_geometry()
    else:
        swidth, sheight, x0, y0 = _parse_tk_geometry(TKtopGUI(parent).geometry())

    if xoff is None:
        xoff = 0
    if yoff is None:
        yoff = 0

    _width = width
    if width is None:
        _width = win.winfo_reqwidth()

    _height = height
    if height is None:
        _height = win.winfo_reqheight()

    x = (swidth - _width) // 2 + x0 + xoff
    y = (sheight - _height) // 2 + y0 + yoff

    if not allow_negative:
        x = np.max([xoff, x])
        y = np.max([yoff, y])

    if width or height:
        win.geometry('%dx%d+%d+%d' % (_width, _height, x, y))
    else:
        win.geometry('+%d+%d' % (x, y))


# ----------------
# extra Tk widgets
# ---------------
class HelpTip(object):
    def __init__(self, widget=None, move=False):
        self.widget = widget
        self.tipwindow = None
        self.mode = 0
        self.location = None
        self.text = None

    def showtip(self, text, location=None, mode=None, move=True, strip=False):
        "Display text in helptip window"
        self.hidetip()
        if location is not None:
            self.location = location
        if mode is not None:
            self.mode = mode
        if text:
            if strip:
                text = text.strip()
            self.text = text

            if isinstance(text, list):
                text = '\n'.join(text)
            text = wrap(text, 100)

            self.tipwindow = tk.Toplevel(self.widget)
            self.tipwindow.withdraw()
            color = "#ffffe0"

            if self.mode == 1:
                # editor mode
                self.tipwindow.wm_transient(self.widget)
                wmax = np.max([max(np.array([len(line) for line in text.split('\n')])), 50])
                hmax = 11
                self.tipwindow.wm_title('Tip')
                self.widget.winfo_containing(self.widget.winfo_pointerx(), self.widget.winfo_pointery()).unbind("<Leave>")
                self.widget.bind_all("<Escape>", self.hidetip)
                self.widget.unbind_all("<F4>")
            else:
                # view mode
                if platform.system() != 'Darwin':
                    self.tipwindow.wm_overrideredirect(1)
                wmax = np.max(np.array([len(line) for line in text.split('\n')]))
                hmax = np.min([text.count('\n') + 1, 11])
                if hmax > 10 and move:
                    ttk.Label(self.tipwindow, text='Press F4 to detach', font=OMFITfont('bold', -1)).pack()
                    self.widget.winfo_containing(self.widget.winfo_pointerx(), self.widget.winfo_pointery()).bind("<Leave>", self.hidetip)
                self.widget.unbind_all("<Escape>")
                if not move:
                    self.widget.bind_all("<F4>", self.changeMode)

            frm = ttk.Frame(self.tipwindow, relief=tk.SOLID, borderwidth=1)
            frm.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)
            label = tk.Text(frm, width=wmax, height=hmax, undo=tk.TRUE, maxundo=-1, wrap='word')
            label.config(borderwidth=0, font=OMFITfont('normal', 0, 'Courier'))
            label.insert(1.0, text)
            label.delete('end -1 chars', 'end')
            label.grid(column=0, row=0, sticky='nsew')
            sb = ttk.Scrollbar(frm, command=label.yview, orient='vertical')
            if hmax > 10:
                sb.grid(column=1, row=0, sticky='ns')

            if self.mode == 1 and self.location:
                label.insert(1.0, '')
                label.focus_set()

                def onButton(event=None):
                    self.text = label.get(1.0, tk.END)
                    exec((self.location + '=' + repr(self.text)), globals(), locals())
                    self.changeMode()
                    return "break"

                bt = ttk.Button(frm, text='Update variables description <{ctrlCmd()}-Return>', command=onButton)
                bt.grid(column=0, row=1, sticky='nsew')
                label.bind(f'<{ctrlCmd()}-Return>', onButton)
            else:
                label.config(state=tk.DISABLED)

            label['yscrollcommand'] = sb.set
            frm.columnconfigure(0, weight=1)
            frm.rowconfigure(0, weight=1)

            if not move:
                x = self.widget.winfo_pointerx()
                y = self.widget.winfo_pointery()
                self.tipwindow.update_idletasks()

                # place window smack in the middle of the cursor
                self.x = int(x - self.tipwindow.winfo_reqwidth() / 2.0)
                self.y = int(y - self.tipwindow.winfo_reqheight() / 2.0)

                # move window off the edges
                if (self.x + self.tipwindow.winfo_reqwidth()) > screen_geometry()[0]:
                    self.x = self.x - self.tipwindow.winfo_reqwidth()
                if (self.y + self.tipwindow.winfo_reqheight()) > screen_geometry()[1]:
                    self.y = self.y - self.tipwindow.winfo_reqheight()
                if self.x < 0:
                    self.x = 0
                if self.y < 0:
                    self.y = 0
                self.tipwindow.wm_geometry("+%d+%d" % (self.x, self.y))
                self.tipwindow.update_idletasks()

                # place cursor smack in the middle of the window
                x = int(self.x + self.tipwindow.winfo_reqwidth() / 2.0)
                y = int(self.y + self.tipwindow.winfo_reqheight() / 2.0)
                self.tipwindow.event_generate('<Motion>', warp=True, x=x, y=y)
                self.tipwindow.update_idletasks()

                self.tipwindow.deiconify()
                self.tipwindow.bind("<Leave>", self.hidetip)

            else:
                self.movetip()
                self.tipwindow.deiconify()

    def hidetip(self, event=None):
        try:
            self.mode = 0
            self.tipwindow.destroy()
            self.tipwindow = None
        except Exception:
            pass

    def movetip(self, event=None):
        try:
            if self.tipwindow is not None:
                if event is None:
                    x = self.widget.winfo_pointerx()
                    y = self.widget.winfo_pointery()
                else:
                    x = event.x_root
                    y = event.y_root
                self.tipwindow.update_idletasks()
                self.x = x + 10
                self.y = y + 12
                if self.x > screen_geometry()[0] - (self.tipwindow.winfo_reqwidth() + 10):
                    self.x = self.x - self.tipwindow.winfo_reqwidth() - 20
                if self.y > screen_geometry()[1] - (self.tipwindow.winfo_reqheight() + 12):
                    self.y = self.y - self.tipwindow.winfo_reqheight() - 24
                self.tipwindow.wm_geometry("+%d+%d" % (self.x, self.y))
                if self.mode == 0:
                    self.tipwindow.after(25, self.movetip)
        except Exception as _excp:
            warnings.warn(repr(_excp))

    def changeMode(self, event=None):
        self.showtip(self.text, mode=np.mod(self.mode + 1, 2))


helpTip = HelpTip()


class ToolTip(object):
    """
    Tooltip recipe from
    http://www.voidspace.org.uk/python/weblog/arch_d7_2006_07_01.shtml#e387
    """

    @staticmethod
    def createToolTip(widget, text):
        toolTip = ToolTip(widget)

        def enter(event):
            toolTip.showtip(text)

        def leave(event):
            toolTip.hidetip()

        widget.bind('<Enter>', enter)
        widget.bind('<Leave>', leave)

    def __init__(self, widget):
        self.widget = widget
        self.tipwindow = None
        self.id = None
        self.x = self.y = 0

    def showtip(self, text):
        "Display text in tooltip window"
        self.text = text
        if self.tipwindow or not self.text:
            return
        x, y, _, _ = self.widget.bbox("insert")
        x = x + self.widget.winfo_rootx() + 27
        y = y + self.widget.winfo_rooty()
        self.tipwindow = tw = tk.Toplevel(self.widget)
        tw.wm_overrideredirect(1)
        tw.wm_geometry("+%d+%d" % (x, y))
        try:
            # For Mac OS
            tw.tk.call("::tk::unsupported::MacWindowStyle", "style", tw._w, "help", "noActivates")
        except tk.TclError:
            pass
        label = tk.Label(tw, text=self.text, justify=tk.LEFT, background="#ffffe0", relief=tk.SOLID, borderwidth=1)
        label.pack(ipadx=1)

    def hidetip(self):
        tw = self.tipwindow
        self.tipwindow = None
        if tw:
            tw.destroy()


def dialog(message='Are You Sure?', answers=['Yes', 'No'], icon='question', title=None, parent=None, options=None, entries=None, **kw):
    """
    Display a dialog box and wait for user input

    Note:

    * The first answer is highlighted (for users to quickly respond with `<Return>`)

    * The last answer is returned for when users press `<Escape>` or close the dialog

    :param message: the text to be written in the label

    :param answers: list of possible answers

    :param icon: "question", "info", "warning", "error"

    :param title: title of the frame

    :param parent: tkinter window to which the dialog will be set as transient

    :param options: dictionary of True/False options that are displayed as checkbuttons in the dialog

    :param entries: dictionary of string options that are displayed as entries in the dialog

    :return: return the answer chosen by the user (a dictionary if options keyword was passed)
    """

    class _Dialog(object):
        def __init__(self, top, message='Are You Sure?', answers=['Yes', 'No'], icon='question', title=None, options=None, entries=None):
            self.top = top
            self.result = None
            self.options_var = {}
            self.entries_var = {}
            top = ttk.Frame(self.top)
            top.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES, padx=10, pady=10)

            frm = ttk.Frame(top)
            frm.pack(side=tk.TOP, fill=tk.X, expand=tk.YES, padx=0, pady=0)
            if icon in ["question", "info", "warning", "error"]:
                if title is None:
                    title = icon
                OMFITsrc = str(os.path.abspath(os.path.dirname(__file__)))
                img = tk.PhotoImage(file=os.path.join(OMFITsrc, 'extras', 'graphics', icon + ".gif"))
                icon_canvas = tk.Canvas(frm, width=64, height=64)
                icon_canvas.copy_image = img
                icon_canvas.create_image(32, 32, image=icon_canvas.copy_image)
                icon_canvas.pack(side=tk.LEFT, fill=tk.NONE, expand=tk.NO)

            ttk.Label(frm, text=message, justify=tk.LEFT).pack(side=tk.LEFT, fill=tk.X, expand=tk.YES)

            if options is not None:
                for item in list(options.keys()):
                    frm = ttk.Frame(top)
                    frm.pack(side=tk.TOP, fill=tk.X, expand=tk.YES, padx=0, pady=0)
                    self.options_var[item] = var = tk.BooleanVar()
                    tmp = ttk.Checkbutton(frm, variable=var, text=item)
                    tmp.var = var
                    tmp.pack(side=tk.LEFT, padx=5, pady=2)
                    var.set(options[item])

            if entries is not None:
                for item in list(entries.keys()):
                    frm = ttk.Frame(top)
                    frm.pack(side=tk.TOP, fill=tk.X, expand=tk.YES, padx=5, pady=2)
                    self.entries_var[item] = var = tk.StringVar()
                    ttk.Label(frm, text=item + ' ').pack(side=tk.LEFT)
                    tmp = ttk.Entry(frm, textvariable=var)
                    tmp.pack(side=tk.LEFT, fill=tk.X, expand=tk.YES)
                    var.set(entries[item])

            frm = ttk.Frame(top)
            frm.pack(side=tk.TOP, fill=tk.X, expand=tk.YES, padx=0, pady=0)
            for ii, answer in enumerate(answers):
                tmp = ttk.Button(frm, text=answer, command=lambda answer=answer: self.ok(answer))
                tmp.bind('<Return>', lambda event: event.widget.invoke())
                tmp.pack(side=tk.LEFT, fill=tk.X, expand=tk.YES)
                if ii == 0:
                    tmp.focus()
            self.top.bind('<Escape>', lambda event=None, answer=answers[-1]: self.ok(answer))
            self.top.protocol("WM_DELETE_WINDOW", lambda event=None, answer=answers[-1]: self.ok(answer))

            if title is None:
                title = 'message'
            self.top.wm_title(title[0].upper() + title[1:])

        def ok(self, x=None):
            if not len(self.options_var) and not len(self.entries_var):
                self.result = x
            else:
                self.result = [x]
                kw = {}
                for item in list(self.options_var.keys()):
                    kw[item] = bool(self.options_var[item].get())
                for item in list(self.entries_var.keys()):
                    kw[item] = self.entries_var[item].get()
                self.result.append(kw)
            self.top.destroy()

    if parent is None:
        parent = OMFITaux['rootGUI']
    try:
        frm_top = tk.Toplevel(parent)
    except Exception:
        parent = OMFITaux['rootGUI']
        frm_top = tk.Toplevel(parent)
    frm_top.withdraw()
    frm_top.transient(parent)
    d = _Dialog(frm_top, message=message, answers=answers, icon=icon, title=title, options=options, entries=entries)
    frm_top.resizable(False, False)
    tk_center(frm_top, parent)
    frm_top.deiconify()
    try:
        frm_top.grab_set()
    except tk.TclError:
        pass
    frm_top.update_idletasks()
    frm_top.wait_window(d.top)
    return d.result


def pickTreeLocation(startLocation=None, title='Pick tree location ...', warnExists=True, parent=None):
    """
    Function which opens a blocking GUI to ask user to pick a tree location

    :param startLocation: starting location

    :param title: title to show in the window

    :param warnExists: warn user if tree location is already in use

    :param parent: tkinter window to which the dialog will be set as transient

    :return: the location picked by the user
    """

    def onEscape(event=None):
        top.destroy()

    if parent is None:
        parent = OMFITaux['rootGUI']
    top = tk.Toplevel(parent)
    top.withdraw()
    top.transient(parent)
    frm = ttk.Frame(top)
    frm.pack(side=tk.TOP, expand=tk.YES, fill=tk.X, padx=5, pady=5)
    frm.grid_columnconfigure(1, weight=1)
    top.wm_title(title)
    outcome = [None]

    def onReturn(event=None):
        v2 = newLocation.get()
        try:
            eval(v2)
            if warnExists and 'No' == dialog(
                title='Location already exists', message='Proceed anyways?', answers=['Yes', 'No'], parent=top
            ):
                return
        except Exception:
            pass
        top.destroy()
        outcome[0] = v2

    ttk.Label(frm, text='To: ').grid(row=1, sticky=tk.E)
    newLocation = tk.OneLineText(frm, width=50, percolator=True)
    if startLocation is None:
        newLocation.set(OMFITaux['GUI'].focusRoot)
    else:
        newLocation.set(startLocation)
    tk_center(top, parent)
    e1 = newLocation
    e1.grid(row=1, column=1, sticky=tk.E + tk.W, columnspan=2)
    e1.focus_set()
    top.bind('<Return>', onReturn)
    top.bind('<KP_Enter>', onReturn)
    top.bind('<Escape>', onEscape)
    top.protocol("WM_DELETE_WINDOW", top.destroy)
    top.update_idletasks()
    top.deiconify()
    top.wait_window(top)
    return outcome[0]


stored_passwords = {}


def password_gui(title='Password', parent=None, key=None):
    """
    Present a password dialog box

    :param title: The title for the dialog box

    :param parent: The GUI parent

    :param key: A key for caching the password in this session
    """
    if key and key in stored_passwords:
        return stored_passwords[key]

    def onEscape(event=None):
        top.destroy()

    if parent is None:
        parent = OMFITaux['rootGUI']

    import tkinter as tk
    from tkinter import ttk

    top = tk.Toplevel(parent)
    top.withdraw()
    top.transient(parent)
    frm = ttk.Frame(top)
    frm.pack(side=tk.TOP, expand=tk.YES, fill=tk.X, padx=5, pady=5)
    frm.grid_columnconfigure(1, weight=1)
    top.wm_title(title)
    outcome = [None]

    def onReturn(event=None):
        outcome[0] = pass_phrase.get()
        top.destroy()

    ttk.Label(frm, text='Pass_phrase').grid(row=1, sticky=tk.E)
    pass_phrase = tk.Entry(frm, width=50, show='*')
    tk_center(top, parent)
    pass_phrase.grid(row=1, column=1, sticky=tk.E + tk.W, columnspan=2)
    pass_phrase.focus_set()
    top.bind('<Return>', onReturn)
    top.bind('<KP_Enter>', onReturn)
    top.bind('<Escape>', onEscape)
    top.protocol("WM_DELETE_WINDOW", top.destroy)
    top.update_idletasks()
    top.deiconify()
    top.wait_window(top)
    if key:
        stored_passwords[key] = outcome[0]
    return outcome[0]


# ----------------
# Tk events
# ---------------
class eventBindings(object):
    def __init__(self):
        self.descs = []
        self.widgets = []
        self.events = []
        self.callbacks = []
        self.tags = []
        self.default_desc_event = {}

    def add(self, description, widget, event, callback, tag=None):
        if '_' in description:
            raise ValueError('The saving/restoring of key bindings cannot handle _ in the description')

        # set event bindings for more than one widget
        if description in self.descs:
            ind = self.descs.index(description)
            if tag:
                widget.tag_bind(tag + '_action', self.events[ind], callback)
            else:
                try:
                    widget.bind(self.events[ind], callback)
                except tk.TclError:
                    if 'ISO' in event:
                        pass
                    else:
                        raise
            self.widgets[ind].append(widget)
            return

        # bind
        if tag:
            widget.tag_bind(tag + '_action', event, callback)
            if platform.system() == 'Darwin' and 'Control' in event:
                widget.tag_bind(tag + '_action', re.sub('Control', 'Command', event), callback)
        else:
            try:
                widget.bind(event, callback)
            except tk.TclError:
                if 'ISO' in event:
                    pass
                else:
                    raise
            if platform.system() == 'Darwin' and 'Control' in event:
                widget.bind(re.sub('Control', 'Command', event), callback)

        # add to list
        self.descs.append(description)
        self.widgets.append([widget])
        self.events.append(event)
        self.callbacks.append(callback)
        if tag:
            self.tags.append(tag + '_action')
        else:
            self.tags.append(tag)
        self.default_desc_event[description.replace(' ', '_')] = event

    def remove_widget(self, widget):
        for wi in range(len(self.widgets)):
            if widget in self.widgets[wi]:
                self.widgets[wi].remove(widget)

    def set(self, description, event):
        try:
            ind = self.descs.index(description)
        except ValueError:
            raise ValueError('Unable to set %s because it is not yet bound to a widget' % description)
        if event == self.events[ind]:
            return
        for w in list(self.widgets[ind]):
            try:
                if self.tags[ind]:
                    w.tk.call(w._w, 'tag', 'bind', self.tags[ind], self.events[ind], '')
                    # Doesn't exist -> w.tag_unbind(self.tags[ind],self.events[ind])
                else:
                    w.unbind(self.events[ind])
            except tk.TclError as _excp:
                if 'bad window path name' in str(_excp) or "application has been destroyed" in str(_excp):
                    self.remove_widget(w)
                    continue
            if self.tags[ind]:
                w.tag_bind(self.tags[ind], event, self.callbacks[ind])
            else:
                w.bind(event, self.callbacks[ind])
            self.events[ind] = event

    def get(self, description):
        for di, d in enumerate(self.descs):
            if d == description:
                return self.events[di]

    def print_event_in_bindings(self, w, event, ind):
        bindings = set()
        for cls in w.bindtags():
            bindings |= set(w.bind_class(cls))  # s |= t means: update set s, adding elements from t
        bindings = [s.replace('Key-', '') for s in list(bindings)]
        printi(event in bindings, self.events[ind] in bindings)

    def printAll(self):
        from utils import printi

        for di, d in enumerate(self.descs):
            printi('%s: %s' % (d, self.events[di]))  # ,self.callbacks[di],self.widgets[di]


global_event_bindings = eventBindings()

# ----------------
# import tk elements
# ---------------
from utils_tk import *
import utils_tk

# ----------------
# screen geometry
# ---------------
def _parse_tk_geometry(geom):
    """
    Function that parses string returned by tk geometry()

    :param geom: string in the format width, height, x, y `400x300+30+55`

    :return: tuple with 4 integers for width, height, x, y
    """
    geometry = []
    geometry.append(int(re.sub(r'([0-9]*)x([0-9]*)\+([\-0-9]*)\+([\-0-9]*)', r'\1', geom)))
    geometry.append(int(re.sub(r'([0-9]*)x([0-9]*)\+([\-0-9]*)\+([\-0-9]*)', r'\2', geom)))
    geometry.append(int(re.sub(r'([0-9]*)x([0-9]*)\+([\-0-9]*)\+([\-0-9]*)', r'\3', geom)))
    geometry.append(int(re.sub(r'([0-9]*)x([0-9]*)\+([\-0-9]*)\+([\-0-9]*)', r'\4', geom)))
    return geometry


_screen_geometry = []


def screen_geometry():
    """
    Function returns the screen geometry

    :return: tuple with 4 integers for width, height, x, y
    """
    if not len(_screen_geometry):
        t = tk.Tk()  # new window
        t.withdraw()

        t.attributes("-alpha", 0)
        try:
            t.state('zoomed')
        except Exception:
            t.attributes('-fullscreen', True)
        t.update_idletasks()
        geom = _parse_tk_geometry(t.geometry())
        geom[0] = t.winfo_reqwidth()
        geom[1] = t.winfo_reqheight()

        _screen_geometry.extend(geom)

        screen = [t.winfo_screenwidth(), t.winfo_screenheight()]
        _screen_geometry[0] = screen[0] - _screen_geometry[2]
        _screen_geometry[1] = screen[1] - _screen_geometry[3]

        t.destroy()

    return _screen_geometry


_displays = [0]


def wmctrl():
    """
    This function is useful when plugging in a new display in OSX and the OMFIT window disappear
    To fix the issue, go to the XQuartz application and select the OMFIT window from the menu `Window > OMFIT ...`
    Then press F8 a few times until the OMFIT GUI appears on one of the screens.
    """
    # screen sizes
    with open(os.devnull, 'w') as nul_f:
        child = subprocess.Popen([OMFITsrc + '/../bin/hmscreens', "-info"], stderr=nul_f, stdout=subprocess.PIPE)
    hms_out, std_err = map(b2s, child.communicate())

    # parse output of hmscreens
    screens = []
    for ks, s in enumerate(hms_out.strip().split('\n\n')):
        s = re.sub(r'\}', ']', re.sub(r'\{', '[', s))
        s = re.sub(r'([\w \(\)]+):(.*)\n*', r'"\1":"\2",', s)
        s = re.sub(r'\n', ',', s)
        screens.append(eval('{%s}' % s))
        for item in screens[-1]:
            try:
                screens[-1][item] = eval(screens[-1][item])
                if isinstance(screens[-1][item], list):
                    screens[-1][item] = np.array(screens[-1][item])
            except (TypeError, NameError):
                pass

    # loop through displays
    display = _displays[0] = np.mod(_displays[0] + 1, len(screens))

    # find geometry according to Tk inter
    minx = 0
    maxx = 0
    miny = 0
    maxy = 0
    for s in screens:
        minx = np.min([minx, s['Global Position'][0, 0]])
        miny = np.min([miny, s['Global Position'][0, 1]])
        maxx = np.max([maxx, s['Global Position'][1, 0]])
        maxy = np.max([maxy, s['Global Position'][1, 1]])

    # change window geometry accordingly
    g = [
        screens[display]['Size'][0],
        screens[display]['Size'][1] - 44 * int(screens[display]['Resolution(dpi)'][1] // 72),
        screens[display]['Global Position'][0, 0] - minx,
        screens[display]['Global Position'][1, 1] - maxy - 44 * int(screens[display]['Resolution(dpi)'][1] // 72),
    ]
    g = '%dx%d+%d+%d' % (g[0], g[1], g[2], g[3])
    OMFITaux['rootGUI'].geometry(g)
    printi('Switched to display #%d: %s' % (display, g))
    OMFITaux['rootGUI'].update_idletasks()


def TKtopGUI(item):
    """
    Function to reach the TopLevel tk from a GUI element

    :param item: tk GUI element

    :return: TopLevel tk GUI
    """
    while True:
        try:
            if item.master is None:
                break
            item = item.master
        except Exception:
            return item
    return item


if __name__ == '__main__':

    main = tk.Tk()
    password_gui(parent=main)
    main.mainloop()



from omfit_classes.utils_base import *

framework = False
repo = None


class OMFITobject(object):
    def __init__(self, filename, **kw):

        readOnly = kw.pop('readOnly', False)
        if readOnly:
            kw['readOnly'] = True
        file_type = kw.pop('file_type', 'file')
        if file_type == 'dir':
            kw['file_type'] = 'dir'
        kw.pop('noCopyToCWD', None)  # this is always the case when running the classes outside the framework
        self.OMFITproperties = {}
        self.OMFITproperties.update(kw)

        # NOTE: some functionalities are available only when the OMFIT framework is running
        for item in ['serverPicker', 'remote', 'server', 'tunnel', 's3bucket']:
            if item in kw:
                raise Exception('`%s` functionality is only available within the OMFIT framework' % item)

        self.modifyOriginal = True  # importing classes outside of OMFIT --always-- operates on the original files
        self.readOnly = readOnly  # readonly functionality is supported outside of framework
        self.originalFilename = filename
        self.filename = filename
        self.link = filename
        self.dynaLoad = False

        if filename is None:
            return

        # remove trailing backslash from filename
        filename = filename.rstrip(os.sep)

        # handle comma separated filenames
        if file_type == 'dir':
            fnames = filename.split(',')
            if len(fnames) > 1:
                filename = os.path.split(filename.split(',')[0])[0]

        # create file if it does not exist
        if not os.path.exists(filename) or not len(filename):
            # if a directory was NOT specified...
            if not len(os.path.split(filename)[0]) or not len(filename):
                # an empty string generates a temporary file
                if not len(filename):
                    import tempfile

                    filename = tempfile._get_default_tempdir() + os.sep + file_type + '_' + now("%Y-%m-%d__%H_%M_%S__%f")
                # create file in the current working directory
                if file_type == 'dir':
                    os.makedirs(filename)
                else:
                    open(filename, 'w').close()
            else:
                raise OMFITexception("No such file or directory: '" + filename + "'")

        # set filename attributes
        filename = os.path.realpath(os.path.expandvars(os.path.expanduser(filename)))
        self.originalFilename = filename
        self.filename = filename
        self.link = filename

        # keep track of what classes have been loaded
        from omfit_classes.utils_base import _loaded_classes

        _loaded_classes.add(self.__class__.__name__)

    def save(self):
        """
        The save method is supposed to be overridden by classes which use OMFITobject as a superclass.
        If left as it is this method can detect if .filename was changed and if so, makes a copy from the original .filename (saved in the .link attribute) to the new .filename
        """
        return self._save_by_copy()

    def _save_by_copy(self):
        # if not exists or different
        if not (os.path.exists(self.filename) and os.path.samefile(self.link, self.filename)):

            if not os.path.exists(self.link):
                raise IOError('Missing link file ' + str(self.link))

            # remove existing file/directory if overwriting
            if not os.path.isdir(self.link):
                if os.path.exists(self.filename):
                    import filecmp

                    if filecmp.cmp(self.link, self.filename, shallow=False):
                        self.link = self.filename
                        return
                    else:
                        os.remove(self.filename)
                shutil.copy2(self.link, self.filename)
            else:
                tmp = os.getcwd()
                try:
                    # change working directory to handle possible overwriting
                    # of the current working directory or its parents
                    os.chdir('/')
                    if os.path.exists(self.filename):
                        shutil.rmtree(self.filename)
                    shutil.copytree(self.link, self.filename)
                finally:
                    os.chdir(tmp)

            self.link = self.filename

    def saveas(self, filename, remove_original_file=False, **kw):
        """
        This function changes the ``.filename`` attribute to filename and calls the ``.save()`` method and optionally remove original file.
        NOTE: use .deploy() to save an object to a given directory without changing the .filename that OMFIT uses

        :param filename: new absolute path of the filename.
                         If relative path is specified, then directory of current filename is used as root.
                         An empty filename will skip the save.

        :param remove_original_file: remove original file (forced to `False` if object is readOnly)

        :return: return value of save()
        """
        # empty filename skips the save
        if not filename:
            return

        # if not absolute path then use directory of current filename as root
        if filename.strip()[0] != os.sep:
            return self.saveas(
                os.path.split(os.path.abspath(self.filename))[0] + os.sep + filename, remove_original_file=remove_original_file, **kw
            )

        # save as
        old_filename = self.filename
        try:
            self.filename = filename
            tmp = self.save(**kw)
        except Exception:
            self.filename = old_filename
            raise

        # remove original file
        if remove_original_file and os.path.exists(old_filename) and not self.readOnly:
            if os.path.isdir(old_filename):
                shutil.rmtree(old_filename)
            else:
                os.remove(old_filename)
        return tmp

    def deploy(self, filename='', **kw):
        """
        The deploy method is used to save an OMFITobject to a location without affecting it's .filename nor .link attributes

        :param filename: filename or directory where to deploy file to
        """
        if filename == '' and self.filename:
            filename = os.path.split(self.filename)[1]

        tmpF = self.filename
        if hasattr(self, 'link'):
            tmpL = self.link

        try:
            if filename == '':
                self.filename = os.path.split(self.filename)[1]
            elif os.path.exists(filename) and os.path.isdir(filename):
                self.filename = filename + os.sep + os.path.split(tmpF)[1]
            else:
                self.filename = filename
            self.filename = os.path.abspath(self.filename)

            directory = os.path.split(self.filename)[0]
            if not os.path.exists(directory):
                os.makedirs(directory)

            self.save(**kw)

            return self.filename

        finally:
            self.filename = tmpF
            if hasattr(self, 'link'):
                self.link = tmpL

    def __deepcopy__(self, memo={}):
        """
        This method attempts to copy by pikling.
        If this fails, it will resort to calling the .duplicate() method

        :param memo: dictionary for preventing loops in deepcopy

        :return: copied object
        """
        try:
            return pickle.loads(pickle.dumps(self, pickle.HIGHEST_PROTOCOL))
        except Exception as _excp:
            printt('Failed to pickle: ' + repr(_excp) + '\nFallback on duplicating by file ' + self.filename)
            return self.duplicate(self.filename)

    def close(self):
        """
        This method:

            1. calls the .save() method, if it exists

            2. calls the .clear() method

            3. calls the .__init__() method with the arguments that were originally fed

        The purpose of this method is to unload from memory the typical OMFIT objects
        (which are a combination of SorteDict+OMFITobject classes), and has been added
        based on the considerations outlined in:
        http://deeplearning.net/software/theano/tutorial/python-memory-management.html#internal-memory-management
        http://www.evanjones.ca/memoryallocator/
        """
        if hasattr(self, 'dynaLoad') and self.dynaLoad:
            return
        if hasattr(self, 'save'):
            self.save()
        if hasattr(self, 'clear'):
            self.clear()
        self.__init__(self.filename, noCopyToCWD=True, **self.OMFITproperties)

    def __tree_repr__(self):
        if isinstance(self.filename, str):
            if os.path.isdir(self.filename):
                values = 'DIR: ' + os.path.split(self.filename)[1]
            else:
                values = 'FILE: ' + os.path.split(self.filename)[1] + '    (' + sizeof_fmt(self.filename) + ")"
        else:
            values = 'DIR/FILE: ?'
        return values, []


from omfit_classes.sortedDict import *


def test_classes_main_header():
    """
    Place this funtion after `if '__main__' == __name__:` in classes/omfit_xxx.py files
    This will:
    1. change working directory to temporary folder
    2. select matplotlib PDF backend if running without $DISPLAY variable set
    """
    # change working directory to temporary folder
    os.chdir(OMFITtmpDir)
    print('Testing class folder: ' + OMFITtmpDir)

    # select matplotlib PDF backend if running without $DISPLAY variable set
    if not os.environ.get('DISPLAY', None):
        print('Setting mpl pdf backend')
        if 'matplotlib' in sys.modules:
            from matplotlib import pyplot

            pyplot.switch_backend('pdf')
        else:
            import matplotlib

            matplotlib.use('agg')


MainScratch = scratch = {}
OMFIT = {'scratch': MainScratch}


class defaultVars(dict):
    """
    Mimic behaviour of defaultVars within OMFIT framework by allowing passing of arguments to standalone scripts at command line.
    For example:

    Given test.py:
        >> from omfit_classes.omfit_ascii import OMFITascii, defaultVars
        >> bla = defaultVars(a=1)
        >> a = bla['a']
        >> print(a)

    Allow passing of arguments with:
        python test.py a=2

    Which will print:
        2
    """

    def __init__(self, *args, **kw):
        dict.__init__(self, *args, **kw)

        import argparse

        parser = argparse.ArgumentParser()
        parser.add_argument('scriptArgs', nargs='*')
        args = parser.parse_args()
        for item in args.scriptArgs:
            if '=' not in item:
                print(args.scriptArgs)
                raise ValueError('Must pass argument as key=value')
            tmp = item.split('=')
            key = tmp[0]
            if key not in self:
                raise ValueError('defaultVars argument `%s` not recognized. Valid arguments are: %s' % (key, self.keys()))
            value = eval('='.join(tmp[1:]))
            self[key] = value


# load the MainSettingsNamelistDump.txt file which is basically the user MainSettings
# with all of the dynamic expressions being evaluated.
from omfit_classes import namelist

OMFIT['MainSettings'] = namelist.NamelistName()
_userMainSettingsFilename = OMFITsettingsDir + os.sep + 'MainSettingsNamelistDump.txt'
_skeleton_filename = os.path.split(os.path.abspath(__file__))[0] + os.sep + 'skeleton' + os.sep + 'skeletonMainSettingsNamelist.txt'
if os.path.exists(_userMainSettingsFilename):
    for fn in [_skeleton_filename, _userMainSettingsFilename]:
        if os.path.exists(fn):
            _tmp = namelist.NamelistFile(fn)
            _tmp.load()
            recursiveUpdate(OMFIT['MainSettings'], _tmp)
else:
    # if MainSettingsNamelistDump.txt is not found then skeletonMainSettingsNamelist.txt is loaded
    # this does not have customizations about usersname and such for connections to remote servers
    # To create this file, start OMFIT frameowork and customize MainSettings via File >  Preferences GUI
    OMFIT['MainSettings'].update(namelist.NamelistFile(_skeleton_filename))
SERVER = OMFIT['MainSettings']['SERVER']



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

import numpy as np
from omfit_classes.omfit_ascii import OMFITascii
import fortranformat

__all__ = ['OMFITaccomeEquilibrium']


class OMFITaccomeEquilibrium(SortedDict, OMFITascii):
    r"""
    Class used to interface equilibrium files generated by ACCOME

    :param filename: filename passed to OMFITascii class

    :param \**kw: keyword dictionary passed to OMFITascii class
    """

    def __init__(self, filename, **kw):
        OMFITascii.__init__(self, filename, **kw)
        SortedDict.__init__(self, caseInsensitive=True)
        self.dynaLoad = True

    def load(self):
        def splitter(inv, step=16):
            value = []
            for k in range(len(inv) // step):
                value.append(inv[step * k : step * (k + 1)])
            return value

        with open(self.filename, 'r') as f:
            lines = f.read().split('\n')
        header = lines[0].split()
        self['HEADER'] = ''.join([x.ljust(8) for x in header[:-4]])
        _ = int(header[-4])  # obsolete variable
        # number of points in the horizontal and vertical directions respectively
        self['NR'] = NR = int(header[-3])
        self['NZ'] = NZ = int(header[-2])
        # number of poloidal flux surfaces
        self['NV'] = N = int(header[-1])
        f2020 = fortranformat.FortranRecordWriter('5e16.9')

        block = []
        for k in range(1, 5):
            block.extend(splitter(lines[k]))
        data0d = list(map(float, block))
        name0d = [
            'RBOX',  # the nominal full-width of the rectangle
            'ZBOX',  # the nominal full-height of the rectangle
            'RADMAJ',  # nominal major radius of the plasma axis
            'RBOXDST',  # major radius of the inner edge of the rectangular grid
            'YMIDEQ',  # the vertical shift of the rectangular box in the up-down symmetry plane
            'RAXIS',  # major radius of the magnetic axis
            'ZAXIS',  # vertical height of the magnetic axis
            'PSIMAG',  # poloidal flux function values at the magnetic axis
            'PSILIM',  # poloidal flux function values at the last closed flux surface
            'BTOR',  # toroidal magnetic field at raxis
            'TOTEQD',  # toroidal plasma current
            '_PSIMX1',  # poloidal flux function at the magnetic axis (again!)
            '_PSIMX2',  # obsolete variable
            '_XAX1',  # major radius of the magnetic axes(xax1 is the same as raxis above)
            '_XAX2',  # xax2 is not used (xax2 = xax1)
            '_ZAX1',  # vertical height of the magnetic axes(xax1 is the same as zaxis above)
            '_ZAX2',  # zax2 is not used (zax2 = zax1)
            '_PSISEP',  # obsolete variable
            '_XSEP',  # obsolete variable
            '_ZSEP',  # obsolete variable
        ]
        for name, data in zip(name0d, data0d):
            if not name.startswith('_'):
                self[name] = data

        block = []
        n = 5 + 4 * int(np.ceil(N / 5.0))
        for k in range(5, n):
            block.extend(splitter(lines[k]))
        data1d = np.reshape(np.array(list(map(float, block))), (4, N))
        self['FPSIAR'] = data1d[0, :]  # 1-D array of (R * B_phi) on "NV" equi-spaced points in psi
        self['PRAR'] = data1d[1, :]  # 1-D array of pressure on "NV" equi-spaced points in psi
        self['FFPAR'] = data1d[2, :]  # 1-D array of f-f_prime on "NV" equi-spaced points in psi
        self['PPAR'] = data1d[3, :]  # 1-D array of p-p_prime on "NV" equi-spaced points in psi

        block = []
        for k in range(n, len(lines)):
            block.extend(splitter(lines[k]))
            if len(block) == NR * NZ:
                break
        self['PSI'] = np.reshape(np.array(list(map(float, block))), (NR, NZ))  # 2-D array of psi values on an "NR" by "NZ" grid

        n = n + int(np.ceil(NR * NZ / 5.0))
        block = []
        for k in range(n, len(lines)):
            block.extend(splitter(lines[k]))
        data = np.array(list(map(float, block)))

        self['Q'] = data[:N]  # array of safety factor on "NV" equi-spaced points in psi

        r, z = np.reshape(data[N:], (-1, 2)).T
        index = np.where((r == r[0]) & (z == z[0]))[0][1] + 1
        self['RBBBS'] = r[:index]
        self['ZBBBS'] = z[:index]
        self['RLIM'] = r[index:]
        self['ZLIM'] = z[index:]


############################################
if '__main__' == __name__:
    test_classes_main_header()

    tmp = OMFITaccomeEquilibrium(OMFITsrc + '/../samples/accome_equilibrium.dat')
    tmp.load()
